'''Common settings shared by the camera server/client
'''


# Hostname (or IP address) of the server. This is the address that the
# client tries to connect to
SERVER_HOSTNAME = '127.0.0.1'
CAMERA_PORT = 50075

# Voltage input output server/client
VIO_PORT = 50085

